from pyneuroutils.transforms.transform import transform
import numpy as np

class sample_1d(transform):
    def __init__(self, n,window,seed=None, **kwargs):
        super().__init__()
        self.n = n
        self.seed = seed
        self.window = window


    def __call__(self,x):
        if self.seed is not None:
            np.random.seed(self.seed)

        if self.window>=x.shape[-1]:
            raise Exception("InvalidInputDimensins: {} vs window size: {}".format(x.shape,self.window))

        if x.ndim==2 and x.shape[0]==1: #shape (1,Nsamp)
            y = np.zeros((self.n,self.window))
            for i in range(self.n):
                idx = np.random.randint(x.shape[-1]-self.window-1)
                y[i,:] = x[0,idx:idx+self.window]
        else:
            raise Exception("InvalidInputDimensins: {}".format(x.shape))

        return y


class sample_spectrogram(transform):
    def __init__(self,n,window,seed=None):
        super().__init__()
        self.window=window
        self.n = n
        self.seed = seed

    def __call__(self, x):
        if self.seed is not None:
            np.random.seed(self.seed)

        if x.ndim != 3:
            raise Exception("InvalidInputDimensins: {}".format(x.shape))

        B = self.n
        CH = x.shape[0]
        H = x.shape[1]
        W = self.window

        y = np.zeros((B,CH,H,W))
        for i in range(self.n):
            idx = np.random.randint(x.shape[-1] - self.window - 1)
            y[i,:,:,:] = x[:,:, idx:idx + self.window]
        return y


from pyneuroutils.transforms.transform import transform
import numpy as np

class segment_1d(transform):
    def __init__(self, window,overlap, **kwargs):
        super().__init__()
        self.window = window
        self.overlap = overlap


    def __call__(self,x):
        if self.window>=x.shape[-1]:
            raise Exception("InvalidInputDimensins: {} vs window size: {}".format(x.shape,self.window))

        if not (x.ndim==2 and x.shape[0]==1):
            #shape (1,Nsamp)
            raise Exception("InvalidInputDimensins: {}".format(x.shape))

        n = int(np.floor((x.shape[-1] - self.window)/(self.window-self.overlap)))+1

        y = np.zeros((n,self.window))
        idx = 0
        for i in range(n):
            y[i,:] = x[0,idx:idx+self.window]
            idx += self.window - self.overlap



        return y



from pyneuroutils.transforms.transform import transform


class crop(transform):
    def __init__(self, dim_0=None, dim_1=None,dim_2=None):
        super().__init__()
        self.dim_0 = dim_0
        self.dim_1 = dim_1
        self.dim_2 = dim_2

    def __call__(self,x):

        if self.dim_0 is not None:
            x = x[self.dim_0[0]:self.dim_0[1],...]

        if self.dim_1 is not None:
            x = x[:,self.dim_1[0]:self.dim_1[1],...]

        if self.dim_2 is not None:
            x = x[:,:,self.dim_2[0]:self.dim_2[1],...]


        return x




class transform(object):
    def __init__(self):
        pass

    def __call__(self,x):
        raise NotImplementedError

    def __repr__(self):
        x = "{}(".format(self.__class__.__name__)

        for key,value in self.__dict__.items():
            if key[0] is not "_":
                x+= "{} = {}, ".format(key,value)
        x=x[:-2]
        x+=')'
        return x




import scipy.signal as sig
from pyneuroutils.transforms.transform import transform

class filtfilt(transform):
    def __init__(self, n, wn, btype='low', **kwargs):
        super().__init__()
        self.n = n
        self.wn = wn
        self.btype = btype
        self._b,self._a = sig.butter(n,wn,btype)

    def __call__(self,x):
        if x.ndim != 2:
            raise Exception("InvalidInputDimensins: {}".format(x.shape))
        return sig.filtfilt(self._b,self._a,x)




from pyneuroutils.transforms.transform import transform
import numpy as np

class diff(transform):
    def __init__(self, axis, **kwargs):
        super().__init__()
        self.axis = axis

    def __call__(self,x):
        if x.ndim != 2:
            raise Exception("InvalidInputDimensins: {}".format(x.shape))
        return np.diff(x,axis=self.axis)



import scipy.signal as sig
from pyneuroutils.transforms.transform import transform


class spectrogram(transform):
    def __init__(self, fs, nperseg, noverlap, nfft, **kwargs):
        super().__init__()
        self.fs = fs
        self.nperseg = nperseg
        self.noverlap = noverlap
        self.nfft = nfft

    def __call__(self, x):
        _,_,y = sig.spectrogram(x,fs=self.fs,nperseg=self.nperseg,noverlap =self.noverlap,nfft=self.nfft)
        return y



import scipy.stats as stats
from pyneuroutils.transforms.transform import transform


class zscore(transform):
    def __init__(self, axis=-1, **kwargs):
        super().__init__()
        self.transform = self.__class__.__name__
        self.axis = axis

    def __call__(self, x):
        return stats.zscore(x,axis=self.axis)


import pickle

class compose:
    """Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            img = t(img)
        return img

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string

    def dumps(self):
        return pickle.dumps(self,fix_imports=True)

    @staticmethod
    def loads(bytes_object):
        return pickle.loads(bytes_object)






import numpy as np
import zmq
import time
import sys
import os
import threading
from pyneuroutils.transforms import *
def timer_connect_error():
    raise Exception("Connection Error")


def timer_recv_error():
    raise Exception("Recv Error")


class MEFD_client(object):
    def __init__(self, ip, wait=60, verbose=False):
        self.ip = ip
        self.wait = wait
        self.verbose = verbose
        self.context = None
        if self.verbose:
            print("client-init", os.getpid())

    def connect(self, port,test_connection=True):
        """
        This should be called in a new spawned process, not in main process -> where __init__ is called
        :param port:
        :return:
        """
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.poll = zmq.Poller()
        self.socket.connect("tcp://{}:{}".format(self.ip, port))
        self.poll.register(self.socket, zmq.POLLIN)
        if self.verbose:
            print('client-connect', os.getpid(), port)
        if test_connection:
            self.socket.send_pyobj({'task': "GET_STATUS"})
            q = self.socket.recv_json()
            print(q)

    def request(self,**kwargs):
        self.socket.send_pyobj(kwargs)
        socks = dict(self.poll.poll(self.wait * 1000))
        if socks.get(self.socket) == zmq.POLLIN:
            response = self.socket.recv_json()
            return response
        else:
            self.socket.setsockopt(zmq.LINGER, 0)
            self.socket.close()
            self.poll.unregister(self.socket)
            raise Exception("Server not responding")


    def request_data(self, path, channel, password, start=None, stop=None, **kwargs):
        WORK = {'path': path,
               'channel': channel,
               'start': start,
               'stop': stop,
               'password': password}
        for key,value in kwargs.items():
            WORK[key] = value
        #self.socket.send_json(WORK)
        self.socket.send_pyobj(WORK)

        socks = dict(self.poll.poll(self.wait * 1000))
        if socks.get(self.socket) == zmq.POLLIN:
            response = self.recv_array()
            return response
        else:
            self.socket.setsockopt(zmq.LINGER, 0)
            self.socket.close()
            self.poll.unregister(self.socket)
            raise Exception("Server not responding")

    def recv_array(self, flags=0, copy=True, track=False):
        """recv a numpy array"""
        md = self.socket.recv_json(flags=flags, )
        if 'exception' in md:
            err = "SERVER: "+md['ip']+":"+md['port']+" ERROR: "+md['exception']
            raise Exception(err)
        msg = self.socket.recv(flags=flags, copy=copy, track=track)
        buf = memoryview(msg)
        A = np.frombuffer(buf, dtype=md['dtype'])
        return A.reshape(md['shape'])



if __name__ == "__main__":
    for i in range(10):
        try:
            t0 = time.time()
            client = MEFD_client(ip='10.144.10.73', wait=10)
            client.connect(port=54321)
            response = client.request(task="GET_STATUS")

            transform =compose([filtfilt(n=3,wn=0.1,btype='low'),
                                sample_1d(n=10,window=5000),
                                zscore(axis=-1)])

            data = client.request_data(path='/mnt/Helium/data/fnusa/seeg-040/Easrec_sciexp-seeg040_151001-1120.mefd',
                                    channel='A1',
                                    password='bemena',
                                    start=None,
                                    stop=None,
                                    transform=transform,
                                    task="GET_DATA")
            stop = 1
        except Exception as exc:
            print(exc)
        finally:
            print(time.time() - t0)
    print("OK")
